import pytest
from web3 import Web3

from multiproof.bytes import to_hex
from multiproof.standard import (LeafValue, StandardMerkleTree,
                                 StandardMerkleTreeData)

ZERO_BYTES = bytearray(32)
ZERO = to_hex(ZERO_BYTES)


def make_tree(s: str, sort_leaves: bool = True) -> tuple[list[list[str]], StandardMerkleTree]:
    l = [[x] for x in s]
    tree = StandardMerkleTree.of(l, ['string'], sort_leaves)
    return l, tree


class TestStandartTestCase:

    @pytest.mark.parametrize('sort_leaves', (True, False))
    def test_valid_single_proofs(self, sort_leaves):
        "generates valid single proofs for all leaves"
        _, tree = make_tree('abcdef', sort_leaves)
        tree.validate()

        for index, leaf in enumerate(tree.values):
            # getProof internally validates the proof
            proof1 = tree.get_proof(index)
            proof2 = tree.get_proof(leaf.value)
            assert proof1 == proof2  # deep equal
            assert tree.verify_leaf(index, proof1)
            assert tree.verify_leaf(leaf, proof1)
            assert StandardMerkleTree.verify(tree.root, ['string'], leaf.value, proof1)

    @pytest.mark.parametrize('sort_leaves', (True, False))
    def test_invalid_single_proofs(self, sort_leaves):
        "rejects invalid proofs"
        _, tree = make_tree('abcdef', sort_leaves)
        _, other_tree = make_tree('abc', sort_leaves)
        leaf = ['a']
        invalid_proof = other_tree.get_proof(leaf)
        assert not tree.verify_leaf(leaf, invalid_proof)
        assert not StandardMerkleTree.verify(tree.root, ['string'], leaf, invalid_proof)

    @pytest.mark.parametrize('sort_leaves', (True, False))
    def test_valid_multiproofs(self, sort_leaves):
        "generates valid multiproofs"
        values, tree = make_tree('abcdef', sort_leaves)
        tree.validate()

        for ids in [[], [0, 1], [0, 1, 5], [1, 3, 4, 5], [0, 2, 4, 5], [0, 1, 2, 3, 4, 5]]:
            # getProof internally validates the proof
            proof1 = tree.get_multi_proof(ids)
            proof2 = tree.get_multi_proof([values[i] for i in ids])
            assert proof1 == proof2  # deep equal
            pf = tree.get_multi_proof(ids)
            assert tree.verify_multi_proof_leaf(proof1)
            assert StandardMerkleTree.verify_multi_proof(tree.root, ['string'], pf)

    @pytest.mark.parametrize('sort_leaves', (True, False))
    def test_invalid_multiproofs(self, sort_leaves):
        "reject invalid multiproofs"
        _, tree = make_tree('abcdef', sort_leaves)
        _, other_tree = make_tree('abc', sort_leaves)
        leaves = [['a'], ['b'], ['c']]
        multi_proof = other_tree.get_multi_proof(leaves)
        assert not tree.verify_multi_proof_leaf(multi_proof)
        assert not StandardMerkleTree.verify_multi_proof(tree.root, ['string'], multi_proof)

    @pytest.mark.parametrize('sort_leaves', (True, False))
    def test_dump_and_load(self, sort_leaves):
        _, tree = make_tree('abcdef', sort_leaves)
        tree2 = StandardMerkleTree.load(tree.dump())
        tree2.validate()

        assert tree2.leaf_encoding == tree.leaf_encoding
        assert tree2.values == tree.values
        assert tree2.tree == tree.tree

    @pytest.mark.parametrize('sort_leaves', (True, False))
    def test_out_of_bonds(self, sort_leaves):
        _, tree = make_tree('a', sort_leaves)
        with pytest.raises(Exception) as context:
            tree.get_proof(1)
            assert 'Index out of bounds' in context.exception

    def test_reject_unrecognized_tree(self):
        with pytest.raises(Exception) as context:
            StandardMerkleTree.load(
                StandardMerkleTreeData(
                    tree=[], values=[], leaf_encoding=['uint256'], format='nonstandard'
                )
            )
            assert "Unknown format 'nonstandard'" in context.exception

    def test_reject_malformed(self):
        with pytest.raises(Exception) as context:
            tree1 = StandardMerkleTree.load(
                StandardMerkleTreeData(
                    format='standard-v1',
                    tree=[ZERO],
                    values=[LeafValue(value=['0'], tree_index=0)],
                    leaf_encoding=['uint256'],
                )
            )
            tree1.get_proof(0)
            assert "Merkle tree does not contain the expected value" in context.exception

        with pytest.raises(Exception) as context:
            tree2 = StandardMerkleTree.load(
                StandardMerkleTreeData(
                    format='standard-v1',
                    tree=[ZERO, ZERO, to_hex(Web3.keccak(Web3.keccak(ZERO_BYTES)))],
                    values=[LeafValue(value=['0'], tree_index=2)],
                    leaf_encoding=['uint256'],
                )
            )
            tree2.get_proof(0)
            assert "Unable to prove value" in context.exception

    def test_render_tree(self):
        "generates valid multiproofs"
        _, tree = make_tree('a')

        expected = '''0) 9c15a6a0eaeed500fd9eed4cbeab71f797cefcc67bfd46683e4d2e6ff7f06d1c'''
        assert str(tree) == expected

        _, tree = make_tree('ab')
        expected = '''
0) fa914d99a18dc32d9725b3ef1c50426deb40ec8d0885dac8edcc5bfd6d030016
├─ 1) 9c15a6a0eaeed500fd9eed4cbeab71f797cefcc67bfd46683e4d2e6ff7f06d1c
└─ 2) 19ba6c6333e0e9a15bf67523e0676e2f23eb8e574092552d5e888c64a4bb3681
        '''.strip()
        assert str(tree) == expected

        _, tree = make_tree('abc')

        expected = '''
0) f2129b5a697531ef818f644564a6552b35c549722385bc52aa7fe46c0b5f46b1
├─ 1) fa914d99a18dc32d9725b3ef1c50426deb40ec8d0885dac8edcc5bfd6d030016
│  ├─ 3) 9c15a6a0eaeed500fd9eed4cbeab71f797cefcc67bfd46683e4d2e6ff7f06d1c
│  └─ 4) 19ba6c6333e0e9a15bf67523e0676e2f23eb8e574092552d5e888c64a4bb3681
└─ 2) 9cf5a63718145ba968a01c1d557020181c5b252f665cf7386d370eddb176517b
    '''.strip()
        assert str(tree) == expected



from typing import cast

from eth_typing import HexStr


def has_hex_prefix(hex_string: str | HexStr) -> bool:
    return hex_string.startswith("0x")


def add_hex_prefix(hex_string: str) -> HexStr:
    if not has_hex_prefix(hex_string):
        return HexStr("0x" + hex_string)
    return cast(HexStr, hex_string)


def remove_hex_prefix(hex_string: HexStr) -> str:
    if has_hex_prefix(hex_string):
        return hex_string[len("0x"):]

    return hex_string


def to_hex(b: bytes) -> HexStr:
    return add_hex_prefix(b.hex())


def hex_to_bytes(hex_string: HexStr) -> bytes:
    return bytes.fromhex(remove_hex_prefix(hex_string))


def compare_bytes(a: bytes, b: bytes) -> int:
    n = min(len(a), len(b))
    for i in range(n):
        if a[i] != b[i]:
            return a[i] - b[i]

    return len(a) - len(b)


def equals_bytes(a: bytes, b: bytes) -> bool:
    if len(a) != len(b):
        return False
    for i, e in enumerate(a):
        if e != b[i]:
            return False
    return True


def concat_bytes(a: bytes, b: bytes) -> bytes:
    return a + b



import math
from dataclasses import dataclass
from itertools import pairwise
from typing import Any

from web3 import Web3

from multiproof.bytes import compare_bytes, concat_bytes, equals_bytes


@dataclass
class CoreMultiProof:
    leaves: list[bytes]
    proof: list[bytes]
    proof_flags: list[bool]


def hash_pair(a: bytes, b: bytes) -> bytes:
    if compare_bytes(a, b) < 0:
        return Web3.keccak(concat_bytes(a, b))
    # pylint: disable=arguments-out-of-order
    return Web3.keccak(concat_bytes(b, a))


def left_child_index(i: int) -> int:
    return 2 * i + 1


def right_child_index(i: int) -> int:
    return 2 * i + 2


def parent_index(i: int) -> int:
    if i > 0:
        return math.floor((i - 1) / 2)
    raise ValueError('Root has no parent')


def sibling_index(i: int) -> int:
    if i > 0:
        return i - (-1) ** (i % 2)
    raise ValueError('Root has no siblings')


def is_tree_node(tree: list[bytes], i: int) -> bool:
    return 0 <= i < len(tree)


def is_internal_node(tree: list[bytes], i: int) -> bool:
    return is_tree_node(tree, left_child_index(i))


def is_leaf_node(tree: list[bytes], i: int) -> bool:
    return is_tree_node(tree, i) and not is_internal_node(tree, i)


def is_valid_merkle_node(node: bytes) -> bool:
    return len(node) == 32


def check_tree_node(tree: list[bytes], i: int) -> None:
    if not is_tree_node(tree, i):
        raise ValueError("Index is not in tree")


def check_internal_node(tree: list[bytes], i: int) -> None:
    if not is_internal_node(tree, i):
        raise ValueError("Index is not an internal tree node")


def check_leaf_node(tree: list[bytes], i: int) -> None:
    if not is_leaf_node(tree, i):
        raise ValueError("Index is not a leaf")


def check_valid_merkle_node(node: bytes) -> None:
    if not is_valid_merkle_node(node):
        raise ValueError("Merkle tree nodes must be byte array of length 32")


def make_merkle_tree(leaves: list[bytes]) -> list[bytes]:
    for leaf in leaves:
        check_valid_merkle_node(leaf)

    if len(leaves) == 0:
        raise ValueError("Expected non-zero number of leaves")

    tree: list[bytes] = [b''] * (2 * len(leaves) - 1)

    for index, leaf in enumerate(leaves):
        tree[len(tree) - 1 - index] = leaf

    for i in range(len(tree) - 1 - len(leaves), -1, -1):
        tree[i] = hash_pair(
            tree[left_child_index(i)],
            tree[right_child_index(i)],
        )
    return tree


def get_proof(tree: list[bytes], index: int) -> list[bytes]:
    check_leaf_node(tree, index)

    proof = []
    while index > 0:
        proof.append(tree[sibling_index(index)])
        index = parent_index(index)

    return proof


def process_proof(leaf: bytes, proof: list[bytes]) -> bytes:
    check_valid_merkle_node(leaf)
    for item in proof:
        check_valid_merkle_node(item)
    result = leaf
    for item in proof:
        result = hash_pair(item, result)
    return result


def get_multi_proof(tree: list[bytes], indices: list[int]) -> CoreMultiProof:
    for index in indices:
        check_leaf_node(tree, index)

    indices = sorted(indices, reverse=True)

    for prev_index, next_index in pairwise(indices):
        if prev_index == next_index:
            raise ValueError("Cannot prove duplicated index")

    stack = indices[:]
    proof = []
    proof_flags = []

    while len(stack) > 0 and stack[0] > 0:
        j = stack.pop(0)  # take from the beginning
        s = sibling_index(j)
        p = parent_index(j)

        if len(stack) and s == stack[0]:
            proof_flags.append(True)
            stack.pop(0)  # consume from the stack
        else:
            proof_flags.append(False)
            proof.append(tree[s])

        stack.append(p)

    if len(indices) == 0:
        proof.append(tree[0])

    return CoreMultiProof(
        leaves=[tree[i] for i in indices],
        proof=proof,
        proof_flags=proof_flags,
    )


def process_multi_proof(multiproof: CoreMultiProof) -> bytes:
    for leaf in multiproof.leaves:
        check_valid_merkle_node(leaf)

    for p in multiproof.proof:
        check_valid_merkle_node(p)

    if len(multiproof.proof) < len([x for x in multiproof.proof_flags if not x]):
        raise ValueError("Invalid multiproof format")

    if len(multiproof.leaves) + len(multiproof.proof) != len(multiproof.proof_flags) + 1:
        raise ValueError("Provided leaves and multiproof are not compatible")

    stack = multiproof.leaves.copy()
    proof = multiproof.proof.copy()

    for flag in multiproof.proof_flags:
        a = stack.pop(0)
        if flag:
            b = stack.pop(0)
        else:
            b = proof.pop(0)

        stack.append(hash_pair(a, b))
    return pop_safe(stack) or proof.pop(0)


def is_valid_merkle_tree(tree: list[bytes]) -> bool:
    for i, node in enumerate(tree):
        if not is_valid_merkle_node(node):
            return False

        l = left_child_index(i)
        r = right_child_index(i)

        if r >= len(tree):
            if l < len(tree):
                return False
        elif not equals_bytes(node, hash_pair(tree[l], tree[r])):
            return False

    return len(tree) > 0


def pop_safe(array: list[Any]) -> Any:
    try:
        return array.pop()
    except IndexError:
        return None



from dataclasses import asdict, dataclass
from functools import cmp_to_key
from typing import Generic, TypeVar

from eth_abi import encode as abi_encode
from eth_typing import HexStr
from web3 import Web3

from multiproof.bytes import compare_bytes, equals_bytes, hex_to_bytes, to_hex
from multiproof.core import (CoreMultiProof, get_multi_proof, get_proof,
                             is_valid_merkle_tree, left_child_index,
                             make_merkle_tree, process_multi_proof,
                             process_proof, right_child_index)
from multiproof.utils import check_bounds

T = TypeVar('T')


@dataclass
class LeafValue(Generic[T]):
    value: T
    tree_index: int


@dataclass
class MultiProof(Generic[T]):
    """
    User-friendly version of multiproof, compare with CoreMultiProof
    """
    leaves: list[T]
    proof: list[HexStr]
    proof_flags: list[bool]


@dataclass
class StandardMerkleTreeData(Generic[T]):
    tree: list[HexStr]
    values: list[LeafValue[T]]
    leaf_encoding: list[str]
    format: str = 'standard-v1'


@dataclass
class HashedValue(Generic[T]):
    value: T
    index: int
    hash: bytes


def standard_leaf_hash(values: T, types: list[str]) -> bytes:
    return Web3.keccak(Web3.keccak(abi_encode(types, values)))  # type: ignore


class StandardMerkleTree(Generic[T]):
    _hash_lookup: dict[HexStr, int]
    tree: list[bytes]
    values: list[LeafValue[T]]
    leaf_encoding: list[str]

    def __init__(self, tree: list[bytes], values: list[LeafValue[T]], leaf_encoding: list[str]):
        self.tree = tree
        self.values = values
        self.leaf_encoding = leaf_encoding
        self._hash_lookup = {}
        for index, leaf_value in enumerate(values):
            self._hash_lookup[to_hex(standard_leaf_hash(leaf_value.value, leaf_encoding))] = index

    @staticmethod
    def of(
            values: list[T], leaf_encoding: list[str], sort_leaves: bool = True
    ) -> 'StandardMerkleTree[T]':
        hashed_values: list[HashedValue[T]] = []
        for index, value in enumerate(values):
            hashed_values.append(
                HashedValue(value=value, index=index, hash=standard_leaf_hash(value, leaf_encoding))
            )
        if sort_leaves:
            hashed_values = sorted(
                hashed_values,
                key=cmp_to_key(lambda a, b: compare_bytes(a.hash, b.hash)),  # type: ignore
            )

        tree = make_merkle_tree([x.hash for x in hashed_values])

        indexed_values = [LeafValue(value=v, tree_index=0) for v in values]

        for leaf_index, hashed_value in enumerate(hashed_values):
            indexed_values[hashed_value.index].tree_index = len(tree) - leaf_index - 1

        return StandardMerkleTree(tree, indexed_values, leaf_encoding)

    @staticmethod
    def load(data: StandardMerkleTreeData[T]) -> 'StandardMerkleTree[T]':
        if data.format != 'standard-v1':
            raise ValueError(f"Unknown format '{data.format}'")
        return StandardMerkleTree(
            [hex_to_bytes(x) for x in data.tree],
            data.values,
            data.leaf_encoding,
        )

    @staticmethod
    def verify(
        root: HexStr, leaf_encoding: list[str], leaf_value: T, proof: list[HexStr]
    ) -> bool:
        leaf_hash = standard_leaf_hash(leaf_value, leaf_encoding)
        implied_root = process_proof(leaf_hash, [hex_to_bytes(x) for x in proof])
        return equals_bytes(implied_root, hex_to_bytes(root))

    @staticmethod
    def verify_multi_proof(root: HexStr, leaf_encoding: list[str], multiproof: MultiProof) -> bool:
        leaf_hashes = [standard_leaf_hash(value, leaf_encoding) for value in multiproof.leaves]
        proof_bytes = [hex_to_bytes(x) for x in multiproof.proof]
        implied_root = process_multi_proof(
            multiproof=CoreMultiProof(
                leaves=leaf_hashes,
                proof=proof_bytes,
                proof_flags=multiproof.proof_flags,
            )
        )

        return equals_bytes(implied_root, hex_to_bytes(root))

    def dump(self) -> StandardMerkleTreeData[T]:
        return StandardMerkleTreeData(
            format='standard-v1',
            tree=[to_hex(v) for v in self.tree],
            values=self.values,
            leaf_encoding=self.leaf_encoding,
        )

    def to_json(self) -> dict:
        return asdict(self.dump())

    @staticmethod
    def from_json(data: dict) -> 'StandardMerkleTree[T]':
        tree_data = StandardMerkleTreeData(
            tree=data['tree'],
            values=[LeafValue(**item) for item in data['values']],
            leaf_encoding=data['leaf_encoding'],
            format=data.get('format', 'standard-v1'),
        )
        return StandardMerkleTree.load(tree_data)

    @property
    def root(self) -> HexStr:
        return to_hex(self.tree[0])

    def validate(self) -> None:
        for i in range(len(self.values)):
            self._validate_value(i)

        if not is_valid_merkle_tree(self.tree):
            raise ValueError("Merkle tree is invalid")

    def leaf_hash(self, leaf: T) -> HexStr:
        return to_hex(standard_leaf_hash(leaf, self.leaf_encoding))

    def leaf_lookup(self, leaf: T) -> int:
        v = self._hash_lookup[self.leaf_hash(leaf)]
        if v is None:
            raise ValueError("Leaf is not in tree")
        return v

    def get_proof(self, leaf: T | int) -> list[HexStr]:
        # input validity
        value_index: int = leaf  # type: ignore
        if not isinstance(leaf, int):
            value_index = self.leaf_lookup(leaf)
        self._validate_value(value_index)

        # rebuild tree index and generate proof
        tree_index = self.values[value_index].tree_index
        proof = get_proof(self.tree, tree_index)

        # check proof
        thee_hash = self.tree[tree_index]
        implied_root = process_proof(thee_hash, proof)

        if not equals_bytes(implied_root, self.tree[0]):
            raise ValueError("Unable to prove value")

        return [to_hex(p) for p in proof]

    def get_multi_proof(self, leaves: list[int] | list[T]) -> MultiProof:
        # input validity
        value_indices: list[int] = []
        for leaf in leaves:
            if isinstance(leaf, int):
                value_indices.append(leaf)
            else:
                value_indices.append(self.leaf_lookup(leaf))

        for value in value_indices:
            self._validate_value(value)

        # rebuild tree indices and generate proof
        indices = [self.values[i].tree_index for i in value_indices]
        proof = get_multi_proof(self.tree, indices)

        # check proof
        implied_root = process_multi_proof(proof)
        if not equals_bytes(implied_root, self.tree[0]):
            raise ValueError('Unable to prove values')

        # return multiproof in hex format
        return MultiProof(
            leaves=[self.values[self._hash_lookup[to_hex(hash)]].value for hash in proof.leaves],
            proof=[to_hex(x) for x in proof.proof],
            proof_flags=proof.proof_flags,
        )

    def verify_leaf(self, leaf: int, proof: list[HexStr]) -> bool:
        return self._verify_leaf(self._get_leaf_hash(leaf), [hex_to_bytes(p) for p in proof])

    def _verify_leaf(self, leaf_hash: bytes, proof: list[bytes]) -> bool:
        implied_root = process_proof(leaf_hash, proof)
        return equals_bytes(implied_root, self.tree[0])

    def verify_multi_proof_leaf(self, multiproof: MultiProof) -> bool:
        return self._verify_multi_proof_leaf(
            CoreMultiProof(
                leaves=[self._get_leaf_hash(leaf) for leaf in multiproof.leaves],
                proof=[hex_to_bytes(proof) for proof in multiproof.proof],
                proof_flags=multiproof.proof_flags,
            )
        )

    def _verify_multi_proof_leaf(self, multi_proof: CoreMultiProof) -> bool:
        implied_root = process_multi_proof(multi_proof)
        return equals_bytes(implied_root, self.tree[0])

    def _validate_value(self, value_index: int) -> bytes:
        check_bounds(self.values, value_index)
        leaf = self.values[value_index]
        check_bounds(self.tree, leaf.tree_index)
        leaf_hash = standard_leaf_hash(leaf.value, self.leaf_encoding)

        if not equals_bytes(leaf_hash, self.tree[leaf.tree_index]):
            raise ValueError("Merkle tree does not contain the expected value")
        return leaf_hash

    def _get_leaf_hash(self, leaf: int) -> bytes:
        if isinstance(leaf, int):
            return self._validate_value(leaf)
        if isinstance(leaf, LeafValue):
            return standard_leaf_hash(leaf.value, self.leaf_encoding)
        return standard_leaf_hash(leaf, self.leaf_encoding)

    def __str__(self):
        if len(self.tree) == 0:
            raise ValueError("Expected non-zero number of nodes")

        stack: list = [[0, []]]
        lines: list = []

        while len(stack) > 0:
            i, path = stack.pop()
            s = ''

            if len(path):
                s += ''.join([['   ', '│  '][p] for p in path[:-1]]) + ['└─ ', '├─ '][path[-1]]
            s += str(i) + ') ' + to_hex(self.tree[i])[2:]

            lines.append(s)
            if right_child_index(i) < len(self.tree):
                stack.append([right_child_index(i), path + [0]])
                stack.append([left_child_index(i), path + [1]])

        return '\n'.join(lines)
